import plotly.io as pio

pio.templates.default = "none"

from .layout import add_night_blocks, \
					get_night_block_subplot_id, \
					plot_event_blocks

from .signals import plot_scalogram, \
					 plot_scalogram_stack

name = "vivplots"



import pandas as pd
from datetime import datetime, timedelta

import plotly.graph_objs as go

NIGHT_BLOCK_YMAX = 1
NIGHT_BLOCK_YMIN = -NIGHT_BLOCK_YMAX
NIGHT_BLOCK_COLOR = '#E3E3E3'
NIGHT_BLOCK_AXIS_NAME = "night_blocks_axis"
NIGHT_BLOCK_AXIS_MULTIPLIER = 1000


def plot_event_blocks(
    fig, event_blocks, row="all", col="all", fill_color=NIGHT_BLOCK_COLOR, opacity=0.15
):
    try:
        hasattr(fig, "add_trace")
    except:
        raise NameError("Figure does not exist")

    for block in event_blocks:
        fig.add_vrect(
            x0=block[0],
            x1=block[1],
            line_width=0,
            fillcolor=fill_color,
            row=row,
            col=col,
            opacity=opacity,
            layer="below",
        )
    return fig



def add_night_blocks(fig,
                     lights_on_str="06:00:00",
                     lights_off_str="20:00:00",
                     color=NIGHT_BLOCK_COLOR,
                     **kwargs):
    try:
        fig._get_subplot_coordinates()
    except Exception as e:
        print("ERROR: Please generate the Plotly Figure with the " \
              "'make_subplots' function.")
        raise e

    xlims = __extract_xlims(fig)
    timestamps = pd.date_range(xlims[0], xlims[1], freq='D')

    try:
        lights_on_time = datetime.strptime(lights_on_str, "%H:%M:%S")
        lights_off_time = datetime.strptime(lights_off_str, "%H:%M:%S")
    except:
        raise Exception("Could not convert string into datetime object")

    row_col_idxs = __extract_passed_in_rows_and_cols(fig, **kwargs)
    if 'row' in kwargs.keys() and 'col' in kwargs.keys():
        kwargs.pop('row')
        kwargs.pop('col')

    for row, col in row_col_idxs:

        if lights_on_time < lights_off_time:
            
            fill_time_blocks_every_day(
                fig=fig,
                timestamps=timestamps, 
                lights_on_time=lights_off_time, 
                lights_off_time=datetime.strptime('23:59:59', '%H:%M:%S'),
                row=row,
                col=col,
                color=color,
                **kwargs
            )
            fill_time_blocks_every_day(
                fig=fig,
                timestamps=timestamps,
                lights_on_time=datetime.strptime('00:00:00', '%H:%M:%S'),
                lights_off_time=lights_on_time,
                row=row,
                col=col,
                color=color,
                **kwargs
            )
            
        else: 
            
            fill_time_blocks_every_day(
                fig=fig,
                timestamps=timestamps, 
                lights_on_time=lights_off_time, 
                lights_off_time=lights_on_time,
                row=row,
                col=col,
                color=color,
                **kwargs
            )

        __overlay_night_block_axis_layout_with_subplot(fig, row, col)



def __extract_xlims(fig):
    minx = None
    maxx = None

    for trace in fig.data:

        if (trace.name is not None) and (NIGHT_BLOCK_AXIS_NAME in trace.name):
            continue

        trace_min = pd.to_datetime(min(trace.x))
        trace_max = pd.to_datetime(max(trace.x))

        if minx == None:
            minx = trace_min
        if maxx == None:
            maxx = trace_max

        if trace_min < minx:
            minx = trace_min

        if trace_max > maxx:
            maxx = trace_max

    return (minx, maxx)



def __extract_passed_in_rows_and_cols(fig, **kwargs):

    if 'row' in kwargs.keys() and 'col' in kwargs.keys():

        row_col_idxs= [(kwargs.get('row'), kwargs.get('col'))]

    else:

        row_col_idxs = fig._get_subplot_coordinates()

    return row_col_idxs



def fill_time_blocks_every_day(fig,
                               timestamps,
                               lights_on_time,
                               lights_off_time,
                               row,
                               col,
                               color=NIGHT_BLOCK_COLOR,
                               **kwargs):
    
    xs = []
    ys = []

    for date in pd.date_range(timestamps.min().date(), timestamps.max().date()):

        start_time = datetime.combine(date, lights_on_time.time())
        stop_time = datetime.combine(date, lights_off_time.time())

        __append_night_block_points_to_xs_and_ys(start_time=start_time,
                                                 stop_time=stop_time,
                                                 xs=xs,
                                                 ys=ys)

    fig.add_trace(
        go.Scatter(
            x=xs,
            y=ys,
            hoverinfo="skip",
            name=f"{NIGHT_BLOCK_AXIS_NAME}{get_night_block_subplot_id(row,col)}",
            showlegend=False,
            mode="none",
            fill="toself",
            fillcolor=color,
            opacity=0.25,
        ),
        **kwargs
    )


def get_night_block_subplot_id(row, col):
    return int(f"{row}{col}") * NIGHT_BLOCK_AXIS_MULTIPLIER


def __append_night_block_points_to_xs_and_ys(start_time, stop_time, xs, ys):
    '''
    In order to plot large vertical night blocks on the plot, go.Scatter()
    requires the input xs and ys for each point to be in a (seemingly) convoluted
    order. The reason we are doing it this way is because passing in a full list
    of points to go.Scatter() is SO MUCH faster iterating through each date and
    calling fig.add_vrect().

    In the following two blocks of appends, each row of xs corresponds to
    the same row of ys. For example, the first xs row, xs.append(start_time),
    corresponds to the ys row, ys.append(NIGHT_BLOCK_YMIN).

    The whole block of xs and ys works together to build up a standalone 
    rectangle that is later filled with the JavaScript using `fill` parameter in 
    the go.Scatter() function. 
    '''
    xs.append(start_time) # bottom left x-coordinate of night block
    xs.append(start_time) # top left x-coordinate of night block
    xs.append(stop_time)  # top right x-coordinate of night block
    xs.append(stop_time)  # bottom right x-coordinate of night block
    xs.append(start_time) # bottom left x-coordinate of night block
    xs.append(None) # force a null point so there is no line between night blocks

    ys.append(NIGHT_BLOCK_YMIN) # bottom left y-coordinate of night block
    ys.append(NIGHT_BLOCK_YMAX) # top left y-coordinate of night block
    ys.append(NIGHT_BLOCK_YMAX) # top right y-coordinate of night block
    ys.append(NIGHT_BLOCK_YMIN) # bottom right y-coordinate of night block
    ys.append(NIGHT_BLOCK_YMIN) # bottom left y-coordinate of night block
    ys.append(None) # force a null point so there is no line between night blocks



def __overlay_night_block_axis_layout_with_subplot(fig, row, col):

    subplot = fig.get_subplot(row=row, col=col)

    xaxis = subplot.yaxis.anchor
    yaxis = subplot.xaxis.anchor

    for data in fig.data:
        if data.name == f"{NIGHT_BLOCK_AXIS_NAME}{get_night_block_subplot_id(row,col)}":

            yaxis_number = get_night_block_subplot_id(row,col)

            if data.xaxis is None:

                data.update(yaxis=f"y{yaxis_number}",xaxis=xaxis)

                fig.layout[f"yaxis{yaxis_number}"] =dict(
                        tickfont=dict(
                            color="rgba(0,0,0,0)"
                        ),
                        gridcolor="rgba(0,0,0,0)",
                        overlaying=yaxis,
                        side="left",
                        fixedrange=True,
                        zeroline=False,
                        range=[0.2, 0.8]
                    )
    




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    'numpy==1.19.5',
    'pandas==1.1.5',
    'plotly==5.4.0'
]

setuptools.setup(
    name="vivplots",
    version="0.0.6",
    author="Thomas Meacham",
    author_email="thomas.meacham@vivent.ch",
    description="Common plotting modules for Vivent",
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vivent-sarl/vivplots",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)


import pandas as pd
import numpy as np
import logging
import sys

from plotly.subplots import make_subplots
import plotly.graph_objects as go

logging.basicConfig(
    stream=sys.stdout, format="%(levelname)s:%(asctime)s:%(funcName)s:%(lineno)s: %(message)s"
)
logger = logging.getLogger()
logger.setLevel(logging.INFO)


def plot_scalogram_stack(
        fig, coefficients_list, times_array, scales_freq_list, yaxis="period", start_row=1
):
    """
    Using plotly, add to a figure a stack of subplots containing scalogram heatmaps.
    The function can plot at most 4 scalograms. The position and size of the colorbars are hard
    coded for plotting 1 to 4 scalograms and the figure margins are hardcoded for a good display.

    :param fig: figure to add the scalogram to
    :param coefficients_list: list of 2D arrays containing the wavelet coefficients
    :param times_array: 1D array of timestamps used to compute the wavelet coefficients
        (one of the dimensions of the wavelet coefficients)
    :param scales_freq_list: list of 1D arrays with the frequencies used to compute the wavelet
        coefficients (one of the dimensions of the wavelet coefficients)
    :param yaxis: string to chose the information to display on the y axis ('period' or 'frequency')
    :param start_row: which subplot to add the stack of scalograms to
    :return: plotly figure with the additional subplots of heatmaps
    """
    try:
        hasattr(fig, 'add_trace')
    except:
        raise NameError('Figure does not exist')

    n = len(coefficients_list)
    if start_row == 1:
        fig = make_subplots(
            rows=n + 1,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.001,
            horizontal_spacing=0.01,
        )
        fig.add_trace(
            go.Scatter(
                x=times_array,
                y=[-1, 1],
                mode="lines",
                showlegend=False,
                opacity=0,
                hoverinfo="skip",
            ),
            row=1,
            col=1,
        )
        fig.update_xaxes(zeroline=False, visible=False, row=1, col=1)
        fig.update_yaxes(
            visible=False,
            showticklabels=False,
            row=1,
            col=1,
        )
        start_row = 2
        logger.debug("Added a dummy plot at the top to preserve layout")

    n_plots = n + (start_row - 1)
    colorbar_params = [[0.15, 0.01], [0.18, 0.01], [0.2, 0.02], [0.4, 0.15], [0.8, -0.4]]
    colorbar_loc = np.linspace(0.9, 0.1, n_plots) + colorbar_params[-n_plots][1]
    for j in range(n_plots, 1, -1):
        scales_freq = scales_freq_list[n_plots - j]  # Hz

        fig = plot_scalogram(
            fig,
            coefficients_list[n_plots - j],
            times_array,
            scales_freq,
            colorbar_len=colorbar_params[-n_plots][0],
            colorbar_loc=colorbar_loc[j - 1],
            yaxis=yaxis,
            row=j,
            n_plots=n_plots,
        )
        logger.info(f"Added Scalogram #{n_plots-(j-1)} of {n} to the stack")

    return fig


def plot_scalogram(
        fig,
        coefficients_list,
        times_array,
        scales_freq,
        colorbar_len,
        colorbar_loc,
        yaxis="period",
        row=1,
        n_plots=1,
):
    scales_freq_str_list = [f"{x:.4e}" for x in scales_freq]  # Hz
    scales_period = 1 / scales_freq  # in seconds
    scales_period_str_list = []
    for s in scales_period:
        mm, ss = divmod(s, 60)
        hh, mm = divmod(mm, 60)
        scales_period_str_list.append(f"{hh:2.0f}h:{mm:2.0f}m:{ss:2.02f}s")

    scales_period_matrix = np.array([scales_period_str_list] * times_array.shape[0]).T
    scales_frequency_matrix = np.array([scales_freq_str_list] * times_array.shape[0]).T

    logger.info("Finished preparing the plotting data and layout")

    if yaxis == "frequency":
        fig = append_scalogram(
            fig,
            np.nan_to_num(coefficients_list),
            times_array,
            scales_freq_str_list,
            scales_period_matrix,
            "<b>Period: %{customdata}</b><br>" + "<i>Frequency: %{y:.4e} [Hz]</i><br>",
            colorbar_len,
            colorbar_loc,
            yaxis_title="Frequency [Hz]",
            row=row,
            )
    else:
        fig = append_scalogram(
            fig,
            np.nan_to_num(coefficients_list),
            times_array,
            scales_period_str_list,
            scales_frequency_matrix,
            "<b>Period: %{y}</b><br>" + "<i>Frequency: %{customdata:.4e} [Hz]</i><br>",
            colorbar_len,
            colorbar_loc,
            yaxis_title="Periods",
            row=row,
            )

    fig.update_xaxes(
        title="Date",
        row=n_plots,
        col=1,
    )
    fig.update_layout(
        hovermode="x unified",
        hoverlabel={
            "namelength": 0,
            "bgcolor": "rgba(255, 255, 255, 0.65)",
            "font": {"color": "black"},
            "bordercolor": "rgba(255, 255, 255, 0.65)",
        },
    )
    fig.update_layout(
        title="Continuous Wavelet Transform Amplitude Spectrum",
        margin=dict(l=145, r=10, t=80, b=80),
        overwrite=True,
    )
    return fig


def append_scalogram(
        fig,
        z_matrix,
        x_array,
        y_array,
        customdata_matrix,
        hovertemplate,
        colorbar_len,
        colorbar_loc,
        yaxis_title="",
        row=1,
):
    try:
        hasattr(fig, 'append_trace')
    except:
        raise NameError('Figure does not exist')

    fig.append_trace(
        go.Heatmap(
            z=z_matrix,
            x=x_array,
            y=y_array,
            customdata=customdata_matrix,
            hovertemplate=hovertemplate + "CWT Amplitude: %{z:.5f}<extra></extra>",
            colorbar=dict(len=colorbar_len, y=colorbar_loc, xpad=4),
            colorscale="viridis",
        ),
        row=row,
        col=1,
    )
    fig.update_yaxes(
        title=yaxis_title,
        type="category",
        ticks="outside",
        nticks=5,
        showspikes=True,
        spikesnap="cursor",
        spikemode="across",
        spikecolor="rgba(255, 255, 255, 0.65)",
        spikethickness=0.1,
        row=row,
        col=1,
    )
    fig.update_xaxes(
        showspikes=True,
        spikesnap="cursor",
        spikemode="across",
        spikecolor="rgba(255, 255, 255, 0.65)",
        spikethickness=0.1,
        row=row,
        col=1,
    )
    return fig
